__author__ = "Julián Arenas-Guerrero"
__credits__ = ["Julián Arenas-Guerrero"]

__license__ = "Apache-2.0"
__maintainer__ = "Julián Arenas-Guerrero"
__email__ = "arenas.guerrero.julian@outlook.com"


import logging
import pandas as pd

from sqlalchemy import create_engine
from sqlalchemy.pool import NullPool


SQL_RDF_DATATYPE = {
    'INTEGER': 'http://www.w3.org/2001/XMLSchema#integer',
    'INT': 'http://www.w3.org/2001/XMLSchema#integer',
    'SMALLINT': 'http://www.w3.org/2001/XMLSchema#integer',
    'DECIMAL': 'http://www.w3.org/2001/XMLSchema#decimal',
    'NUMERIC': 'http://www.w3.org/2001/XMLSchema#decimal',
    'FLOAT': 'http://www.w3.org/2001/XMLSchema#double',
    'REAL': 'http://www.w3.org/2001/XMLSchema#double',
    'DOUBLE': 'http://www.w3.org/2001/XMLSchema#double',
    'BOOL': 'http://www.w3.org/2001/XMLSchema#boolean',
    'TINYINT': 'http://www.w3.org/2001/XMLSchema#boolean',
    'BOOLEAN': 'http://www.w3.org/2001/XMLSchema#boolean',
    'DATE': 'http://www.w3.org/2001/XMLSchema#date',
    'TIME': 'http://www.w3.org/2001/XMLSchema#time',
    'DATETIME': 'http://www.w3.org/2001/XMLSchema#',
    'TIMESTAMP': 'http://www.w3.org/2001/XMLSchema#dateTime',
    'BINARY': 'http://www.w3.org/2001/XMLSchema#hexBinary',
    'VARBINARY': 'http://www.w3.org/2001/XMLSchema#hexBinary',
    'BIT': 'http://www.w3.org/2001/XMLSchema#hexBinary',
    'YEAR': 'http://www.w3.org/2001/XMLSchema#integer'
}


def _replace_query_enclosing_characters(sql_query, db_dialect):
    db_dialect = db_dialect.upper()
    dialect_sql_query = ''

    if db_dialect in ['MYSQL', 'MARIADB']:
        dialect_sql_query = sql_query   # the query already uses backticks as enclosed characters
    elif db_dialect == 'MSSQL':
        # replace backticks with square brackets
        square_brackets = ['[', ']']
        num_enclosing_char = 0
        for char in sql_query:
            if char == '`':
                dialect_sql_query = dialect_sql_query + square_brackets[num_enclosing_char % 2]
                num_enclosing_char += 1
            else:
                dialect_sql_query = dialect_sql_query + char
    else:
        # replace backticks with double quotes
        dialect_sql_query = sql_query.replace('`', '"')

    return dialect_sql_query


def relational_db_connection(config, source_name):
    db_connection = create_engine(config.get_database_url(source_name), poolclass=NullPool)
    db_dialect = db_connection.dialect.name.upper()

    return db_connection, db_dialect


def get_column_datatype(config, source_name, table_name, column_name):
    db_connection, db_dialect = relational_db_connection(config, source_name)

    sql_query = "SELECT `data_type` FROM `information_schema`.`columns` WHERE `table_name`='" + table_name + \
                "' AND `column_name`='" + column_name + "'"

    sql_query = _replace_query_enclosing_characters(sql_query, db_dialect)

    try:
        query_results_df = pd.read_sql(sql_query, con=db_connection)
    except:
        raise Exception('Query [' + sql_query + '] has failed to execute.')

    data_type = ''
    if 'data_type' in query_results_df.columns and len(query_results_df) == 1:
        data_type = query_results_df['data_type'][0]
    elif 'DATA_TYPE' in query_results_df.columns and len(query_results_df) == 1:
        data_type = query_results_df['DATA_TYPE'][0]

    if data_type.upper() in SQL_RDF_DATATYPE:
        return SQL_RDF_DATATYPE[data_type.upper()]
    else:
        return 'http://www.w3.org/2001/XMLSchema#string'


def build_sql_join_query(mapping_rule, parent_triples_map_rule, references, parent_references):
    if pd.notna(mapping_rule['query']) or pd.notna(parent_triples_map_rule['query']):
        # This is because additional work is needed to use proper aliases within the rml:query provided
        raise Exception(
            'Pushing down SQL joins is not supported for mapping rules using rml:query instead of rr:tablename')

    child_query = build_sql_subquery(mapping_rule, references, 'child_')
    parent_query = build_sql_subquery(parent_triples_map_rule, parent_references, 'parent_')

    query = 'SELECT * FROM (' + child_query + ') AS `child`, (' + parent_query + ') AS `parent` WHERE '
    for key, join_condition in eval(mapping_rule['join_conditions']).items():
        query = query + '`child`.`child_' + join_condition['child_value'] + \
                '`=`parent`.`parent_' + join_condition['parent_value'] + '` AND '
    query = query[:-5]

    logging.debug('SQL query for mapping rule `' + str(mapping_rule['id']) + '`: [' + query + ']')

    return query


def build_sql_subquery(mapping_rule, references, alias):
    if pd.notna(mapping_rule['query']):
        query = mapping_rule['query']
    elif len(references) > 0:
        query = 'SELECT '
        if len(references) > 0:
            for reference in references:
                query = query + '`' + reference + '` AS `' + alias + reference + '`, '
            query = query[:-2] + ' FROM `' + mapping_rule['tablename'] + '` WHERE '
            for reference in references:
                query = query + '`' + reference + '` IS NOT NULL AND '
            query = query[:-5]
    else:
        query = None

    return query


def build_sql_query(config, mapping_rule, references):
    """
    Build a query for MYSQL using backticks '`' as enclosing character. This character will later be replaced with the
    one corresponding one to the dialect that applies.
    """

    if pd.notna(mapping_rule['query']):
        query = mapping_rule['query']
    elif len(references) > 0:
        query = 'SELECT '
        if config.push_down_sql_distincts():
            query = query + 'DISTINCT '
        for reference in references:
            query = query + '`' + reference + '`, '
        query = query[:-2] + ' FROM `' + mapping_rule['tablename'] + '` WHERE '
        for reference in references:
            query = query + '`' + reference + '` IS NOT NULL AND '
        query = query[:-5]
    else:
        query = None

    if query is not None:
        logging.debug('SQL query for mapping rule `' + str(mapping_rule['id']) + '`: [' + query + ']')

    return query


def get_sql_data(config, mapping_rule, references, parent_triples_map_rule=None, parent_references=None):
    db_connection, db_dialect = relational_db_connection(config, mapping_rule['source_name'])

    if parent_triples_map_rule is not None:
        sql_query = build_sql_join_query(mapping_rule, parent_triples_map_rule, references, parent_references)
    else:
        sql_query = build_sql_query(config, mapping_rule, references)

    sql_query = _replace_query_enclosing_characters(sql_query, db_dialect)

    result_chunks = pd.read_sql(sql_query,
                                con=db_connection,
                                chunksize=config.get_chunksize(),
                                coerce_float=False)

    return result_chunks



__author__ = "Julián Arenas-Guerrero"
__credits__ = ["Julián Arenas-Guerrero"]

__license__ = "Apache-2.0"
__maintainer__ = "Julián Arenas-Guerrero"
__email__ = "arenas.guerrero.julian@outlook.com"


import logging
import pandas as pd
import multiprocessing as mp

from itertools import permutations

from ..constants import *
from ..utils import get_mapping_rule_from_triples_map_id


def get_invariant_of_template(template):
    """
    Retrieves the part of the template before the first reference. This part of the template does not depend on
    reference and therefore is invariable. If the template has no references, it is an invalid template, and an
    exception is thrown.
    """

    template_for_splitting = template.replace('\\{', AUXILIAR_UNIQUE_REPLACING_STRING)
    if '{' in template_for_splitting:
        invariant_of_template = template_for_splitting.split('{')[0]
        invariant_of_template = invariant_of_template.replace(AUXILIAR_UNIQUE_REPLACING_STRING, '\\{')
    else:
        # no references were found in the template, and therefore the template is invalid
        raise Exception("Invalid template `" + template + "`. No pairs of unescaped curly braces were found.")

    return invariant_of_template


def _generate_maximal_partition_for_a_position_ordering(mappings_df, position_ordering):
    """
    Generates a maximal mapping partition for a position ordering.
    """

    for position in position_ordering:
        current_global_group = mappings_df.at[0, 'mapping_partition']  # to simulate a groupby
        current_group = 0
        current_invariant = AUXILIAR_UNIQUE_REPLACING_STRING
        current_literal_type = AUXILIAR_UNIQUE_REPLACING_STRING

        # ---------------------------- SUBJECT ----------------------------

        if position == 'S':
            mappings_df.sort_values(by=['mapping_partition', 'subject_invariant'], inplace=True, ascending=True)

            for i, mapping_rule in mappings_df.iterrows():
                if current_global_group != mapping_rule['mapping_partition']:
                    current_group = 0
                    current_invariant = AUXILIAR_UNIQUE_REPLACING_STRING
                    current_global_group = mapping_rule['mapping_partition']

                if mapping_rule['subject_termtype'] == R2RML_BLANK_NODE:
                    mappings_df.at[i, 'mapping_partition'] = mappings_df.at[i, 'mapping_partition'] + '-0'
                elif mapping_rule['subject_invariant'].startswith(current_invariant):
                    mappings_df.at[i, 'mapping_partition'] = mappings_df.at[i, 'mapping_partition'] + '-' + str(
                        current_group)
                else:
                    current_group += 1
                    current_invariant = mapping_rule['subject_invariant']
                    mappings_df.at[i, 'mapping_partition'] = mappings_df.at[i, 'mapping_partition'] + '-' + str(
                        current_group)

        # ---------------------------- PREDICATE ----------------------------

        if position == 'P':
            mappings_df.sort_values(by=['mapping_partition', 'predicate_invariant'], inplace=True, ascending=True)

            # if all predicates are constant terms we can use full string comparison instead of startswith
            enforce_invariant_non_subset = mappings_df['predicate_constant'].notna().all()

            for i, mapping_rule in mappings_df.iterrows():
                if current_global_group != mapping_rule['mapping_partition']:
                    current_group = 0
                    current_invariant = AUXILIAR_UNIQUE_REPLACING_STRING
                    current_global_group = mapping_rule['mapping_partition']

                if enforce_invariant_non_subset and mapping_rule['predicate_invariant'] == current_invariant:
                    mappings_df.at[i, 'mapping_partition'] = mappings_df.at[i, 'mapping_partition'] + '-' + str(
                        current_group)
                elif not enforce_invariant_non_subset and mapping_rule['predicate_invariant'].startswith(
                        current_invariant):
                    mappings_df.at[i, 'mapping_partition'] = mappings_df.at[i, 'mapping_partition'] + '-' + str(
                        current_group)
                else:
                    current_group += 1
                    current_invariant = mapping_rule['predicate_invariant']
                    mappings_df.at[i, 'mapping_partition'] = mappings_df.at[i, 'mapping_partition'] + '-' + str(
                        current_group)

        # ---------------------------- OBJECT ----------------------------

        if position == 'O':
            mappings_df.sort_values(by=['mapping_partition', 'object_termtype', 'literal_type', 'object_invariant'],
                                    inplace=True, ascending=True)

            for i, mapping_rule in mappings_df.iterrows():
                if current_global_group != mapping_rule['mapping_partition']:
                    current_group = 0
                    current_invariant = AUXILIAR_UNIQUE_REPLACING_STRING
                    current_global_group = mapping_rule['mapping_partition']

                if mapping_rule['object_termtype'] == R2RML_BLANK_NODE:
                    mappings_df.at[i, 'mapping_partition'] = mappings_df.at[i, 'mapping_partition'] + '-0'
                elif mapping_rule['object_termtype'] == R2RML_LITERAL:
                    if mapping_rule['literal_type'] != current_literal_type:
                        current_group += 1
                        current_literal_type = mapping_rule['literal_type']
                    mappings_df.at[i, 'mapping_partition'] = mappings_df.at[i, 'mapping_partition'] + '-' + str(
                        current_group)
                elif mapping_rule['object_invariant'].startswith(current_invariant):
                    mappings_df.at[i, 'mapping_partition'] = mappings_df.at[i, 'mapping_partition'] + '-' + str(
                        current_group)
                else:
                    current_group += 1
                    current_invariant = mapping_rule['object_invariant']
                    mappings_df.at[i, 'mapping_partition'] = mappings_df.at[i, 'mapping_partition'] + '-' + str(
                        current_group)

        # ---------------------------- GRAPH ----------------------------

        if position == 'G':
            mappings_df.sort_values(by=['mapping_partition', 'graph_invariant'], inplace=True, ascending=True)

            # if all graph are constant terms we can use full string comparison instead of startswith
            enforce_invariant_non_subset = mappings_df['graph_constant'].notna().all()

            for i, mapping_rule in mappings_df.iterrows():
                if current_global_group != mapping_rule['mapping_partition']:
                    current_group = 0
                    current_invariant = AUXILIAR_UNIQUE_REPLACING_STRING
                    current_global_group = mapping_rule['mapping_partition']

                if enforce_invariant_non_subset and mapping_rule['graph_invariant'] == current_invariant:
                    mappings_df.at[i, 'mapping_partition'] = mappings_df.at[i, 'mapping_partition'] + '-' + str(
                        current_group)
                elif not enforce_invariant_non_subset and mapping_rule['graph_invariant'].startswith(current_invariant):
                    mappings_df.at[i, 'mapping_partition'] = mappings_df.at[i, 'mapping_partition'] + '-' + str(
                        current_group)
                else:
                    current_group += 1
                    current_invariant = mapping_rule['graph_invariant']
                    mappings_df.at[i, 'mapping_partition'] = mappings_df.at[i, 'mapping_partition'] + '-' + str(
                        current_group)

    return mappings_df


class MappingPartitioner:

    def __init__(self, mappings_df, config):
        self.mappings_df = mappings_df
        self.config = config

    def __str__(self):
        return str(self.mappings_df)

    def __repr__(self):
        return repr(self.mappings_df)

    def __len__(self):
        return len(self.mappings_df)

    def partition_mappings(self):
        """
        Generates the mapping partition for the mapping rules in the input DataFrame based on the provided mapping
        partitioning criteria. A new column in the DataFrame is added indicating the mapping partition assigned to every
        mapping rule.
        """

        self.mappings_df = self.mappings_df.reset_index(drop=True)

        if self.config.get_mapping_partition() == PARTIAL_AGGREGATIONS_PARTITIONING:
            self._get_term_invariants()
            self._generate_partial_aggregations_partition()
        elif self.config.get_mapping_partition() == MAXIMAL_PARTITIONING:
            self._get_term_invariants()
            self._generate_maximal_partition()
        elif self.config.get_mapping_partition() in NO_PARTITIONING:
            # assign empty partition
            self.mappings_df['mapping_partition'] = '0-0-0-0'
        else:
            logging.error('Selected mapping partitioning algorithm is not valid.')

        logging.info(
            'Mapping partition with ' + str(len(set(self.mappings_df['mapping_partition']))) + ' groups generated.')
        logging.info('Maximum number of rules within mapping group: ' + str(
            self.mappings_df['mapping_partition'].value_counts()[0]) + '.')

        return self.mappings_df

    def _generate_maximal_partition(self):
        """
        Generates a mapping partition with the maximum number of mapping groups.
        """

        self.mappings_df['literal_type'] = self.mappings_df['object_language'] + self.mappings_df['object_datatype']
        self.mappings_df['mapping_partition'] = ''

        mappings_df = self.mappings_df.copy()

        if mappings_df['predicate_constant'].notna().all():
            logging.debug('All predicate maps are constant-valued, invariant subset is not enforced.')
        if mappings_df['graph_constant'].notna().all():
            logging.debug('All graph maps are constant-valued, invariant subset is not enforced.')

        position_orderings = list(permutations(['S', 'P', 'O', 'G']))

        if self.config.is_multiprocessing_enabled():
            pool = mp.Pool(self.config.get_number_of_processes())
            mapping_partitions_dfs = pool.starmap(_generate_maximal_partition_for_a_position_ordering,
                                                  zip([self.mappings_df.copy()] * len(position_orderings),
                                                      position_orderings))
        else:
            mapping_partitions_dfs = []
            for position_ordering in position_orderings:
                mapping_partitions_dfs.append(
                    _generate_maximal_partition_for_a_position_ordering(self.mappings_df.copy(), position_ordering))

        max_num_groups = -1
        maximal_partition = None
        for mapping_partitions_df in mapping_partitions_dfs:
            if len(set(mapping_partitions_df['mapping_partition'])) > max_num_groups:
                max_num_groups = len(set(mapping_partitions_df['mapping_partition']))
                maximal_partition = mapping_partitions_df

        maximal_partition['mapping_partition'] = maximal_partition['mapping_partition'].str[1:]
        # drop the auxiliary columns that were created just to generate the mapping partition
        maximal_partition.drop([
            'subject_invariant',
            'predicate_invariant',
            'object_invariant',
            'graph_invariant',
            'literal_type'],
            axis=1, inplace=True)

        self.mappings_df = maximal_partition

    def _generate_partial_aggregations_partition(self):
        """
        Generates a mapping partition by independently partitioning by Subject, Predicate, Object and Graph, and
        aggregating this independent partitions.
        """

        # initialize empty mapping partition
        self.mappings_df['subject_partition'] = ''
        self.mappings_df['predicate_partition'] = ''
        self.mappings_df['object_partition'] = ''
        self.mappings_df['graph_partition'] = ''

        self.mappings_df['literal_type'] = self.mappings_df['object_language'] + self.mappings_df['object_datatype']

        # generate partial mapping partition for subjects, predicates, objects and graphs

        # ---------------------------- SUBJECT ----------------------------

        # sort the mapping rules based on the subject invariants. An invariant that starts with another
        # invariant is placed behind in the ordering e.g. http://example.org/term/something and
        # http://example.org/term: http://example.org/term is placed first
        self.mappings_df.sort_values(by=['subject_invariant'], inplace=True, ascending=True)

        current_group = 0
        current_invariant = AUXILIAR_UNIQUE_REPLACING_STRING

        # iterate over the mapping rules and check if the invariant starts with the invariant of the previous rule
        # if it does, then it is in the same mapping partition than the previous mapping rule
        # if it does not, then the mapping rule is in a new mapping partition
        for i, mapping_rule in self.mappings_df.iterrows():
            if mapping_rule['subject_termtype'] == R2RML_BLANK_NODE:
                self.mappings_df.at[i, 'subject_partition'] = '0'
            elif mapping_rule['subject_invariant'].startswith(current_invariant):
                self.mappings_df.at[i, 'subject_partition'] = str(current_group)
            else:
                current_group += 1
                current_invariant = mapping_rule['subject_invariant']
                self.mappings_df.at[i, 'subject_partition'] = str(current_group)

        # ---------------------------- PREDICATE ----------------------------

        self.mappings_df.sort_values(by='predicate_invariant', inplace=True, ascending=True)

        current_group = 0
        current_invariant = AUXILIAR_UNIQUE_REPLACING_STRING

        # if all predicates are constant terms we can use full string comparison instead of startswith
        enforce_invariant_non_subset = self.mappings_df['predicate_constant'].notna().all()
        if enforce_invariant_non_subset:
            logging.debug('All predicate maps are constant-valued, invariant subset is not enforced.')

        for i, mapping_rule in self.mappings_df.iterrows():
            if enforce_invariant_non_subset and mapping_rule['predicate_invariant'] == current_invariant:
                self.mappings_df.at[i, 'predicate_partition'] = str(current_group)
            elif not enforce_invariant_non_subset and mapping_rule['predicate_invariant'].startswith(current_invariant):
                self.mappings_df.at[i, 'predicate_partition'] = str(current_group)
            else:
                current_group += 1
                current_invariant = mapping_rule['predicate_invariant']
                self.mappings_df.at[i, 'predicate_partition'] = str(current_group)

        # ---------------------------- OBJECT ----------------------------

        self.mappings_df.sort_values(by=['object_termtype', 'literal_type', 'object_invariant'],
                                     inplace=True, ascending=True)

        current_group = 0
        current_literal_type = AUXILIAR_UNIQUE_REPLACING_STRING
        current_invariant = AUXILIAR_UNIQUE_REPLACING_STRING

        for i, mapping_rule in self.mappings_df.iterrows():
            if mapping_rule['object_termtype'] == R2RML_BLANK_NODE:
                self.mappings_df.at[i, 'object_partition'] = '0'
            elif mapping_rule['object_termtype'] == R2RML_LITERAL:
                if mapping_rule['literal_type'] != current_literal_type:
                    current_group += 1
                    current_literal_type = mapping_rule['literal_type']
                self.mappings_df.at[i, 'object_partition'] = str(current_group)
            elif mapping_rule['object_invariant'].startswith(current_invariant):
                self.mappings_df.at[i, 'object_partition'] = str(current_group)
            else:
                current_group += 1
                current_invariant = mapping_rule['object_invariant']
                self.mappings_df.at[i, 'object_partition'] = str(current_group)

        # ---------------------------- GRAPH ----------------------------

        self.mappings_df.sort_values(by=['graph_invariant'], inplace=True, ascending=True)

        current_group = 0
        current_invariant = AUXILIAR_UNIQUE_REPLACING_STRING

        # if all graph are constant terms we can use full string comparison instead of startswith
        enforce_invariant_non_subset = self.mappings_df['graph_constant'].notna().all()
        if enforce_invariant_non_subset:
            logging.debug('All graph maps are constant-valued, invariant subset is not enforced.')

        for i, mapping_rule in self.mappings_df.iterrows():
            if enforce_invariant_non_subset and mapping_rule['graph_invariant'] == current_invariant:
                self.mappings_df.at[i, 'graph_partition'] = str(current_group)
            elif not enforce_invariant_non_subset and mapping_rule['graph_invariant'].startswith(current_invariant):
                self.mappings_df.at[i, 'graph_partition'] = str(current_group)
            else:
                current_group += 1
                current_invariant = mapping_rule['graph_invariant']
                self.mappings_df.at[i, 'graph_partition'] = str(current_group)

        # aggregate the independent mapping partition generated for subjects, predicates and graphs to generate the
        # final mapping partition
        self.mappings_df['mapping_partition'] = self.mappings_df['subject_partition'].astype(str) + '-' + \
            self.mappings_df['predicate_partition'].astype(str) + '-' + \
            self.mappings_df['object_partition'].astype(str) + '-' + self.mappings_df['graph_partition'].astype(str)

        # drop the auxiliary columns that were created just to generate the mapping partition
        self.mappings_df.drop([
            'subject_partition',
            'subject_invariant',
            'predicate_partition',
            'predicate_invariant',
            'object_partition',
            'object_invariant',
            'graph_partition',
            'graph_invariant',
            'literal_type'],
            axis=1, inplace=True)

    def _get_term_invariants(self):
        """
        Adds in the input DataFrame new columns for the invariants of mapping rules. Columns for the invariants of
        subjects, predicates and graphs are added, and they are completed based on the provided mapping
        partitioning criteria.
        """

        # initialize empty invariants for all terms
        self.mappings_df['subject_invariant'] = ''
        self.mappings_df['predicate_invariant'] = ''
        self.mappings_df['object_invariant'] = ''
        self.mappings_df['graph_invariant'] = ''

        for i, mapping_rule in self.mappings_df.iterrows():
            # SUBJECT
            if pd.notna(mapping_rule['subject_template']):
                self.mappings_df.at[i, 'subject_invariant'] = \
                    get_invariant_of_template(str(mapping_rule['subject_template']))
            elif pd.notna(mapping_rule['subject_constant']):
                self.mappings_df.at[i, 'subject_invariant'] = str(mapping_rule['subject_constant'])

            # PREDICATE
            if pd.notna(mapping_rule['predicate_constant']):
                self.mappings_df.at[i, 'predicate_invariant'] = str(mapping_rule['predicate_constant'])
            elif pd.notna(mapping_rule['predicate_template']):
                self.mappings_df.at[i, 'predicate_invariant'] = \
                    get_invariant_of_template(str(mapping_rule['predicate_template']))

            # OBJECT
            if pd.notna(mapping_rule['object_constant']):
                self.mappings_df.at[i, 'object_invariant'] = str(mapping_rule['object_constant'])
            elif pd.notna(mapping_rule['object_template']):
                self.mappings_df.at[i, 'object_invariant'] = \
                    get_invariant_of_template(str(mapping_rule['object_template']))
            elif pd.notna(mapping_rule['object_parent_triples_map']):
                # get the invariant for referencing object maps
                parent_mapping_rule = get_mapping_rule_from_triples_map_id(self.mappings_df, mapping_rule[
                    'object_parent_triples_map'])
                if pd.notna(parent_mapping_rule['subject_constant']):
                    self.mappings_df.at[i, 'object_invariant'] = str(parent_mapping_rule['subject_constant'])
                elif pd.notna(parent_mapping_rule['subject_template']):
                    self.mappings_df.at[i, 'object_invariant'] = \
                        get_invariant_of_template(str(parent_mapping_rule['subject_template']))

            # GRAPH
            if pd.notna(mapping_rule['graph_constant']):
                self.mappings_df.at[i, 'graph_invariant'] = str(mapping_rule['graph_constant'])
            elif pd.notna(mapping_rule['graph_template']):
                self.mappings_df.at[i, 'graph_invariant'] = \
                    get_invariant_of_template(str(mapping_rule['graph_template']))



__author__ = "Julián Arenas-Guerrero"
__copyright__ = "Copyright (C) 2020-2021 Julián Arenas-Guerrero"
__credits__ = ["Julián Arenas-Guerrero"]

__license__ = "Apache-2.0"
__maintainer__ = "Julián Arenas-Guerrero"
__email__ = "arenas.guerrero.julian@outlook.com"


import multiprocessing as mp


##############################################################################
########################   ARGUMENTS DEFAULT VALUES   ########################
##############################################################################

DEFAULT_OUTPUT_DIR = 'output'
DEFAULT_OUTPUT_FILE = 'result'
DEFAULT_OUTPUT_FORMAT = 'N-QUADS'
DEFAULT_REMOVE_DUPLICATES = 'yes'
DEFAULT_CLEAN_OUTPUT_DIR = 'yes'
DEFAULT_LOGGING_FILE = ''
DEFAULT_LOGGING_LEVEL = 'INFO'
DEFAULT_INFER_SQL_DATATYPES = 'no'
DEFAULT_NUMBER_OF_PROCESSES = 2 * mp.cpu_count()
DEFAULT_CHUNKSIZE = 100000
DEFAULT_NA_FILTER = 'yes'
DEFAULT_NA_VALUES = ',#N/A,N/A,#N/A N/A,n/a,NA,<NA>,#NA,NULL,null,NaN,nan'
DEFAULT_ONLY_PRINTABLE_CHARACTERS = 'no'
DEFAULT_MATERIALIZE_DEFAULT_GRAPH = 'no'

# DEVELOPMENT OPTIONS
DEFAULT_PUSH_DOWN_SQL_DISTINCTS = 'no'
DEFAULT_PUSH_DOWN_SQL_JOINS = 'no'
DEFAULT_READ_PARSED_MAPPINGS_PATH = ''
DEFAULT_WRITE_PARSED_MAPPINGS_PATH = ''


##############################################################################
########################   MAPPING PARTITION OPTIONS   #######################
##############################################################################

PARTIAL_AGGREGATIONS_PARTITIONING = 'PARTIAL-AGGREGATIONS'
MAXIMAL_PARTITIONING = 'MAXIMAL'
NO_PARTITIONING = ['NO', 'FALSE', 'OFF', '0']


##############################################################################
#########################   DATA SOURCE TYPES   ##############################
##############################################################################

RDB_SOURCE_TYPE = 'RDB'
CSV_SOURCE_TYPE = 'CSV'
TSV_SOURCE_TYPE = 'TSV'
EXCEL_SOURCE_TYPE = 'EXCEL'
PARQUET_SOURCE_TYPE = 'PARQUET'
FEATHER_SOURCE_TYPE = 'FEATHER'
ORC_SOURCE_TYPE = 'ORC'
STATA_SOURCE_TYPE = 'STATA'
SAS_SOURCE_TYPE = 'SAS'
SPSS_SOURCE_TYPE = 'SPSS'

TABULAR_SOURCE_TYPES = [CSV_SOURCE_TYPE, TSV_SOURCE_TYPE, EXCEL_SOURCE_TYPE, PARQUET_SOURCE_TYPE, FEATHER_SOURCE_TYPE,
                        ORC_SOURCE_TYPE, STATA_SOURCE_TYPE, SAS_SOURCE_TYPE, SPSS_SOURCE_TYPE]
DATA_SOURCE_TYPES = [RDB_SOURCE_TYPE] + TABULAR_SOURCE_TYPES


##############################################################################
#########################   VALID ARGUMENTS VALUES   #########################
##############################################################################

VALID_OUTPUT_FORMATS = ['N-TRIPLES', 'N-QUADS']
VALID_PROCESS_START_METHOD = ['DEFAULT', 'SPAWN', 'FORK', 'FORKSERVER']
VALID_LOGGING_LEVEL = ['NOTSET', 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']


##############################################################################
###################   FILE EXTENSIONS FOR OUTPUT FORMATS   ###################
##############################################################################

OUTPUT_FORMAT_FILE_EXTENSION = {
    'N-TRIPLES': '.nt',
    'N-QUADS': '.nq'
}


##############################################################################
###########################   R2RML SPECIFICATION   ##########################
##############################################################################

# classes
R2RML_BASE_TABLE_OR_VIEW_CLASS = 'http://www.w3.org/ns/r2rml#BaseTableOrView'
R2RML_GRAPH_MAP_CLASS = 'http://www.w3.org/ns/r2rml#GraphMap'
R2RML_JOIN_CLASS = 'http://www.w3.org/ns/r2rml#Join'
R2RML_LOGICAL_TABLE_CLASS = 'http://www.w3.org/ns/r2rml#LogicalTable'
R2RML_OBJECT_MAP_CLASS = 'http://www.w3.org/ns/r2rml#ObjectMap'
R2RML_PREDICATE_MAP_CLASS = 'http://www.w3.org/ns/r2rml#PredicateMap'
R2RML_PREDICATE_OBJECT_MAP_CLASS = 'http://www.w3.org/ns/r2rml#PredicateObjectMap'
R2RML_R2RML_VIEW_CLASS = 'http://www.w3.org/ns/r2rml#R2RMLView'
R2RML_REF_OBJECT_MAP_CLASS = 'http://www.w3.org/ns/r2rml#RefObjectMap'
R2RML_SUBJECT_MAP_CLASS = 'http://www.w3.org/ns/r2rml#SubjectMap'
R2RML_TERM_MAP_CLASS = 'http://www.w3.org/ns/r2rml#TermMap'
R2RML_TRIPLES_MAP_CLASS = 'http://www.w3.org/ns/r2rml#TriplesMap'

# properties
R2RML_LOGICAL_TABLE = 'http://www.w3.org/ns/r2rml#logicalTable'
R2RML_TABLE_NAME = 'http://www.w3.org/ns/r2rml#tableName'
R2RML_PARENT_TRIPLES_MAP = 'http://www.w3.org/ns/r2rml#parentTriplesMap'
R2RML_SUBJECT_MAP = 'http://www.w3.org/ns/r2rml#subjectMap'
R2RML_PREDICATE_MAP = 'http://www.w3.org/ns/r2rml#predicateMap'
R2RML_OBJECT_MAP = 'http://www.w3.org/ns/r2rml#objectMap'
R2RML_GRAPH_MAP = 'http://www.w3.org/ns/r2rml#graphMap'
R2RML_SUBJECT_CONSTANT_SHORTCUT = 'http://www.w3.org/ns/r2rml#subject'
R2RML_PREDICATE_CONSTANT_SHORTCUT = 'http://www.w3.org/ns/r2rml#predicate'
R2RML_OBJECT_CONSTANT_SHORTCUT = 'http://www.w3.org/ns/r2rml#object'
R2RML_GRAPH_CONSTANT_SHORTCUT = 'http://www.w3.org/ns/r2rml#graph'
R2RML_PREDICATE_OBJECT_MAP = 'http://www.w3.org/ns/r2rml#predicateObjectMap'
R2RML_CONSTANT = 'http://www.w3.org/ns/r2rml#constant'
R2RML_TEMPLATE = 'http://www.w3.org/ns/r2rml#template'
R2RML_COLUMN = 'http://www.w3.org/ns/r2rml#column'
R2RML_CLASS = 'http://www.w3.org/ns/r2rml#class'
R2RML_CHILD = 'http://www.w3.org/ns/r2rml#child'
R2RML_PARENT = 'http://www.w3.org/ns/r2rml#parent'
R2RML_JOIN_CONDITION = 'http://www.w3.org/ns/r2rml#joinCondition'
R2RML_DATATYPE = 'http://www.w3.org/ns/r2rml#datatype'
R2RML_LANGUAGE = 'http://www.w3.org/ns/r2rml#language'
R2RML_SQL_QUERY = 'http://www.w3.org/ns/r2rml#sqlQuery'
R2RML_SQL_VERSION = 'http://www.w3.org/ns/r2rml#sqlVersion'
R2RML_TERM_TYPE = 'http://www.w3.org/ns/r2rml#termType'

# other
R2RML_DEFAULT_GRAPH = 'http://www.w3.org/ns/r2rml#defaultGraph'
R2RML_IRI = 'http://www.w3.org/ns/r2rml#IRI'
R2RML_LITERAL = 'http://www.w3.org/ns/r2rml#Literal'
R2RML_BLANK_NODE = 'http://www.w3.org/ns/r2rml#BlankNode'
R2RML_SQL2008 = 'http://www.w3.org/ns/r2rml#SQL2008'


##############################################################################
############################   RML SPECIFICATION   ###########################
##############################################################################

QL_NAMESPACE = 'http://semweb.mmlab.be/ns/ql#'
QL_CSV = 'http://semweb.mmlab.be/ns/ql#CSV'
QL_JSON = 'http://semweb.mmlab.be/ns/ql#JSONPath'
QL_XML = 'http://semweb.mmlab.be/ns/ql#XPath'

RML_NAMESPACE = 'http://semweb.mmlab.be/ns/rml#'
RML_LOGICAL_SOURCE = 'http://semweb.mmlab.be/ns/rml#logicalSource'
RML_QUERY = 'http://semweb.mmlab.be/ns/rml#query'
RML_ITERATOR = 'http://semweb.mmlab.be/ns/rml#iterator'
RML_REFERENCE = 'http://semweb.mmlab.be/ns/rml#reference'
RML_REFERENCE_FORMULATION = 'http://semweb.mmlab.be/ns/rml#referenceFormulation'


AUXILIAR_UNIQUE_REPLACING_STRING = 'zzyy_xxww\u200B'



__author__ = "Julián Arenas-Guerrero"
__credits__ = ["Julián Arenas-Guerrero"]

__license__ = "Apache-2.0"
__maintainer__ = "Julián Arenas-Guerrero"
__email__ = "arenas.guerrero.julian@outlook.com"


import re
import os
import shutil
import logging
import rdflib
import time


def configure_logger(logging_level, logging_file):
    """
    Configures the logger. If a logging file is provided, the logs messages are redirected to it, if not they are
    redirected to stdout. Messages are logged according to the provided logging level.
    """

    logging_level_string_to_numeric = {
        'CRITICAL': logging.CRITICAL,
        'ERROR': logging.ERROR,
        'WARNING': logging.WARNING,
        'INFO': logging.INFO,
        'DEBUG': logging.DEBUG,
        'NOTSET': logging.NOTSET,
    }

    if logging_file:
        logging.basicConfig(filename=logging_file,
                            format='%(levelname)s | %(asctime)s | %(message)s',
                            filemode='w',
                            level=logging_level_string_to_numeric[logging_level.upper()])
    else:
        logging.basicConfig(format='%(levelname)s | %(asctime)s | %(message)s',
                            level=logging_level_string_to_numeric[logging_level.upper()])


def get_valid_dir_path(dir_path):
    """
    Checks that a directory exists. If the directory does not exist, it creates the directories in the path.
    """

    dir_path = str(dir_path).strip()
    if not os.path.exists(dir_path):
        os.makedirs(dir_path)

    return dir_path


def get_valid_file_path(file_path):
    """
    Checks that directories in a file path exist. If they do not exist, it creates the directories. Also generates
    a valid file name.
    """

    file_path = str(file_path).strip()
    if not os.path.exists(os.path.dirname(file_path)):
        if os.path.dirname(file_path):
            os.makedirs(os.path.dirname(file_path))

    file_path = os.path.join(os.path.dirname(file_path), get_valid_file_name(os.path.basename(file_path)))

    return file_path


def get_valid_file_name(file_name):
    """
    Generates a valid file name from an input file name.
    """

    file_name = str(file_name).strip()
    file_name.replace(' ', '_')
    file_name = re.sub(r'(?u)[^-\w.]', '', file_name)

    return file_name


def get_repeated_elements_in_list(input_list):
    """
    Finds repeated elements in a list. Returns a list with the repeated elements.
    """

    elem_count = {}
    for elem in input_list:
        if elem in elem_count:
            elem_count[elem] += 1
        else:
            elem_count[elem] = 1

    repeated_elems = []
    for elem, elem_count in elem_count.items():
        if elem_count > 1:
            repeated_elems.append(elem)

    return repeated_elems


def get_subject_maps(mappings_df):
    """
    Retrieves subject maps from mapping rules in the input DataFrame. No repeated subject maps are returned.
    """

    subject_maps_df = mappings_df[[
        'id', 'triples_map_id', 'source_name', 'source_type', 'data_source', 'iterator', 'tablename',
        'query', 'subject_template', 'subject_reference', 'subject_constant', 'subject_termtype']
    ]

    subject_maps_df = subject_maps_df.drop_duplicates()

    return subject_maps_df


def get_references_in_template(template):
    """
    Retrieves all reference identifiers in a template-valued term map. References are returned in order of appearance
    """

    return re.findall('\\{([^}]+)', template)


def triples_to_file(triples, config, mapping_partition=''):
    """
    Writes triples to file.
    """

    f = open(config.get_output_file_path(mapping_partition), 'a')
    for triple in triples:
        f.write(triple + '.\n')
    f.close()


def remove_non_printable_characters(string):
    """
    Eliminates from the input string all the characters that are not printable.
    """

    return ''.join(char for char in string if char.isprintable())


def clean_output_dir(config):
    """
    Removes all files and directories within output_dir in config depending on clean_output_dir parameter. The output
    file, if provided in config, is always deleted.
    """

    output_dir = config.get_output_dir()
    output_file = config.get_output_file()
    if output_file:
        if os.path.exists(os.path.join(output_dir, output_file)):
            # always delete output file, so that generated triples are not appended to it
            os.remove(os.path.join(output_dir, output_file))

    if config.clean_output_dir():
        for obj in os.listdir(output_dir):
            obj_path = os.path.join(output_dir, obj)
            if os.path.isdir(obj_path):
                shutil.rmtree(obj_path)
            else:
                os.remove(obj_path)

        logging.debug('Cleaned output directory.')


def dataframe_columns_to_str(df):
    """
    Converts all the columns in the input dataframe to str.
    """

    return df.astype(str)


def replace_predicates_in_graph(graph, predicate_to_remove, predicate_to_add):
    """
    Replaces the predicates predicate_to_remove in a graph with the predicate predicate_to_add.
    """

    # get the triples with the predicate to be replaced
    r2_rml_sources_query = 'SELECT ?s ?o WHERE {?s <' + predicate_to_remove + '> ?o .}'
    subjects_objects_matched = graph.query(r2_rml_sources_query)

    # for each triple to be replaced add a similar one (same subject and object) but with the new predicate
    for s, o in subjects_objects_matched:
        graph.add((s, rdflib.term.URIRef(predicate_to_add), o))

    # remove all triples in the graph that have the old predicate
    graph.remove((None, rdflib.term.URIRef(predicate_to_remove), None))

    return graph


def get_mapping_rule_from_triples_map_id(mappings, parent_triples_map_id):
    """
    Get the parent triples map of the mapping rule with the given id
    """

    parent_triples_map = mappings[mappings['triples_map_id'] == parent_triples_map_id]

    return parent_triples_map.iloc[0]


def get_delta_time(start_time):
    return "{:.3f}".format((time.time() - start_time))


def add_references_in_join_condition(mapping_rule, references, parent_references):
    references_join, parent_references_join = get_references_in_join_condition(mapping_rule)

    references.update(set(references_join))
    parent_references.update(set(parent_references_join))

    return references, parent_references


def get_references_in_join_condition(mapping_rule):
    references = list()
    parent_references = list()

    join_conditions = eval(mapping_rule['join_conditions'])
    for join_condition in join_conditions.values():
        references.append(join_condition['child_value'])
        parent_references.append(join_condition['parent_value'])

    return references, parent_references


def remove_file_extension(file_path):
    # only removes the last file extension (if there are multiple dots in file_path)
    if len(os.path.splitext(file_path)) > 1:
        return os.path.splitext(file_path)[0]

    return file_path



__author__ = "Julián Arenas-Guerrero"
__credits__ = ["Julián Arenas-Guerrero"]

__license__ = "Apache-2.0"
__maintainer__ = "Julián Arenas-Guerrero"
__email__ = "arenas.guerrero.julian@outlook.com"


from .engine import *
from .args_parser import parse_config



##############################################################################
#######################   MAPPING DATAFRAME COLUMNS   ########################
##############################################################################

MAPPINGS_DATAFRAME_COLUMNS = [
    'source_name', 'triples_map_id', 'data_source', 'object_map', 'ref_form', 'iterator', 'tablename', 'query',
    'subject_template', 'subject_reference', 'subject_constant', 'subject_rdf_class', 'subject_termtype',
    'graph_constant', 'graph_reference', 'graph_template',
    'predicate_constant', 'predicate_template', 'predicate_reference',
    'object_constant', 'object_template', 'object_reference', 'object_termtype', 'object_datatype', 'object_language',
    'object_parent_triples_map', 'join_conditions',
    'predicate_object_graph_constant', 'predicate_object_graph_reference', 'predicate_object_graph_template'
]


##############################################################################
########################   MAPPING PARSING QUERIES   #########################
##############################################################################

MAPPING_PARSING_QUERY = """
    # This query has been reused from SDM-RDFizer (https://github.com/SDM-TIB/SDM-RDFizer). SDM-RDFizer has been
    # developed by members of the Scientific Data Management Group at TIB. Its development has been coordinated and
    # supervised by Maria-Esther Vidal. The implementation has been done by Enrique Iglesias and Guillermo Betancourt
    # under the supervision of David Chaves-Fraga, Samaneh Jozashoori, and Kemele Endris.
    # It has been partially modified by Julián Arenas-Guerrero, PhD student at the Ontology Engineering Group (OEG)
    # in Universidad Politécnica de Madrid (UPM).

    prefix rr: <http://www.w3.org/ns/r2rml#>
    prefix rml: <http://semweb.mmlab.be/ns/rml#>

    SELECT DISTINCT
        ?triples_map_id ?data_source ?ref_form ?iterator ?tablename ?query ?object_map
        ?subject_template ?subject_reference ?subject_constant
        ?subject_rdf_class ?subject_termtype
        ?graph_constant ?graph_reference ?graph_template
        ?predicate_constant ?predicate_template ?predicate_reference
        ?object_constant ?object_template ?object_reference ?object_termtype ?object_datatype ?object_language
        ?object_parent_triples_map
        ?predicate_object_graph_constant ?predicate_object_graph_reference ?predicate_object_graph_template

    WHERE {
        ?triples_map_id rml:logicalSource ?_source .
        OPTIONAL { ?_source rml:source ?data_source . }
        OPTIONAL { ?_source rml:referenceFormulation ?ref_form . }
        OPTIONAL { ?_source rml:iterator ?iterator . }
        OPTIONAL { ?_source rr:tableName ?tablename . }
        OPTIONAL { ?_source rml:query ?query . }

    # Subject -------------------------------------------------------------------------
        OPTIONAL {
            ?triples_map_id rr:subjectMap ?_subject_map .
            OPTIONAL { ?_subject_map rr:template ?subject_template . }
            OPTIONAL { ?_subject_map rml:reference ?subject_reference . }
            OPTIONAL { ?_subject_map rr:constant ?subject_constant . }
            OPTIONAL { ?_subject_map rr:class ?subject_rdf_class . }
            OPTIONAL { ?_subject_map rr:termType ?subject_termtype . }
            OPTIONAL {
                ?_subject_map rr:graphMap ?_graph_structure .
                ?_graph_structure rr:constant ?graph_constant .
            }
            OPTIONAL {
                ?_subject_map rr:graphMap ?_graph_structure .
                ?_graph_structure rr:template ?graph_template .
            }
            OPTIONAL {
                ?_subject_map rr:graphMap ?_graph_structure .
                ?_graph_structure rr:reference ?graph_reference .
            }
        }

    # Predicate -----------------------------------------------------------------------
        OPTIONAL {
            ?triples_map_id rr:predicateObjectMap ?_predicate_object_map .
            OPTIONAL {
                ?_predicate_object_map rr:predicateMap ?_predicate_map .
                ?_predicate_map rr:constant ?predicate_constant .
            }
            OPTIONAL {
                ?_predicate_object_map rr:predicateMap ?_predicate_map .
                ?_predicate_map rr:template ?predicate_template .
            }
            OPTIONAL {
                ?_predicate_object_map rr:predicateMap ?_predicate_map .
                ?_predicate_map rml:reference ?predicate_reference .
            }

    # Object --------------------------------------------------------------------------
            OPTIONAL {
                ?_predicate_object_map rr:objectMap ?object_map .
                ?object_map rr:constant ?object_constant .
                OPTIONAL { ?object_map rr:termType ?object_termtype . }
                OPTIONAL { ?object_map rr:datatype ?object_datatype . }
                OPTIONAL { ?object_map rr:language ?object_language . }
            }
            OPTIONAL {
                ?_predicate_object_map rr:objectMap ?object_map .
                ?object_map rr:template ?object_template .
                OPTIONAL { ?object_map rr:termType ?object_termtype . }
                OPTIONAL { ?object_map rr:datatype ?object_datatype . }
                OPTIONAL { ?object_map rr:language ?object_language . }
            }
            OPTIONAL {
                ?_predicate_object_map rr:objectMap ?object_map .
                ?object_map rml:reference ?object_reference .
                OPTIONAL { ?object_map rr:termType ?object_termtype . }
                OPTIONAL { ?object_map rr:datatype ?object_datatype . }
                OPTIONAL { ?object_map rr:language ?object_language . }
            }
            OPTIONAL {
                ?_predicate_object_map rr:objectMap ?object_map .
                ?object_map rr:parentTriplesMap ?object_parent_triples_map .
                OPTIONAL { ?object_map rr:termType ?object_termtype . }
            }
            OPTIONAL {
                ?_predicate_object_map rr:graphMap ?_graph_structure .
                ?_graph_structure rr:constant ?predicate_object_graph_constant .
            }
            OPTIONAL {
                ?_predicate_object_map rr:graphMap ?_graph_structure .
                ?_graph_structure rr:template ?predicate_object_graph_template .
            }
            OPTIONAL {
                ?_predicate_object_map rr:graphMap ?_graph_structure .
                ?_graph_structure rr:reference ?predicate_object_graph_reference .
            }
        }
    }
"""


JOIN_CONDITION_PARSING_QUERY = """
    prefix rr: <http://www.w3.org/ns/r2rml#>

    SELECT DISTINCT ?object_map ?join_condition ?child_value ?parent_value
    WHERE {
        ?object_map rr:joinCondition ?join_condition .
        ?join_condition rr:child ?child_value;
                        rr:parent ?parent_value.
    }
"""


__author__ = "Julián Arenas-Guerrero"
__credits__ = ["Julián Arenas-Guerrero"]

__license__ = "Apache-2.0"
__maintainer__ = "Julián Arenas-Guerrero"
__email__ = "arenas.guerrero.julian@outlook.com"


from .args_parser import parse_config
from .engine import retrieve_mappings, process_materialization


if __name__ == "__main__":

    config = parse_config()

    mappings = retrieve_mappings(config)
    process_materialization(mappings, config)



__author__ = "Julián Arenas-Guerrero"
__credits__ = ["Julián Arenas-Guerrero"]

__license__ = "Apache-2.0"
__maintainer__ = "Julián Arenas-Guerrero"
__email__ = "arenas.guerrero.julian@outlook.com"


import pandas as pd
import multiprocessing as mp
import logging
import time

from itertools import repeat
from falcon.uri import encode

from .utils import *
from .constants import *
from .data_source.relational_source import get_sql_data
from .data_source.tabular_source import get_table_data


def _get_references_in_mapping_rule(mapping_rule, only_subject_map=False):
    references = []
    if pd.notna(mapping_rule['subject_template']):
        references.extend(get_references_in_template(str(mapping_rule['subject_template'])))
    elif pd.notna(mapping_rule['subject_reference']):
        references.append(str(mapping_rule['subject_reference']))

    if not only_subject_map:
        if pd.notna(mapping_rule['predicate_template']):
            references.extend(get_references_in_template(str(mapping_rule['predicate_template'])))
        elif pd.notna(mapping_rule['predicate_reference']):
            references.append(str(mapping_rule['predicate_reference']))
        if pd.notna(mapping_rule['object_template']):
            references.extend(get_references_in_template(str(mapping_rule['object_template'])))
        elif pd.notna(mapping_rule['object_reference']):
            references.append(str(mapping_rule['object_reference']))
        if pd.notna(mapping_rule['graph_template']):
            references.extend(get_references_in_template(str(mapping_rule['graph_template'])))
        elif pd.notna(mapping_rule['graph_reference']):
            references.append(str(mapping_rule['graph_reference']))

    return set(references)


def _materialize_template(results_df, template, config, columns_alias='', termtype=R2RML_IRI, language_tag='',
                          datatype=''):
    references = get_references_in_template(str(template))

    if str(termtype).strip() == R2RML_LITERAL:
        results_df['triple'] = results_df['triple'] + '"'
    else:
        results_df['triple'] = results_df['triple'] + '<'

    for reference in references:
        results_df['reference_results'] = results_df[columns_alias + reference]

        if config.only_write_printable_characters():
            results_df['reference_results'] = results_df['reference_results'].apply(lambda x: remove_non_printable_characters(encode(x)))

        if str(termtype).strip() == R2RML_IRI:
            # falcon's encode is faster than urllib's quote
            results_df['reference_results'] = results_df['reference_results'].apply(lambda x: encode(x))
        elif str(termtype).strip() == R2RML_LITERAL:
            results_df['reference_results'] = results_df['reference_results'].str.replace('"', '\\"', regex=False).str.replace('\\', '\\\\"', regex=False)

        splitted_template = template.split('{' + reference + '}')
        results_df['triple'] = results_df['triple'] + splitted_template[0] + results_df['reference_results']
        template = str('{' + reference + '}').join(splitted_template[1:])
    if template:
        # add what remains in the template after the last reference
        results_df['triple'] = results_df['triple'] + template

    if str(termtype).strip() == R2RML_LITERAL:
        results_df['triple'] = results_df['triple'] + '"'
        if pd.notna(language_tag):
            results_df['triple'] = results_df['triple'] + '@' + language_tag + ' '
        elif pd.notna(datatype):
            results_df['triple'] = results_df['triple'] + '^^<' + datatype + '> '
        else:
            results_df['triple'] = results_df['triple'] + ' '
    else:
        results_df['triple'] = results_df['triple'] + '> '

    return results_df


def _materialize_reference(results_df, reference, config, columns_alias='', termtype=R2RML_LITERAL,
                           language_tag='', datatype=''):
    results_df['reference_results'] = results_df[columns_alias + str(reference)]

    if config.only_write_printable_characters():
        results_df['reference_results'] = results_df['reference_results'].apply(lambda x: remove_non_printable_characters(encode(x)))

    if str(termtype).strip() == R2RML_IRI:
        # falcon's encode is faster than urllib's quote
        results_df['reference_results'] = results_df['reference_results'].apply(lambda x: encode(x))
        results_df['triple'] = results_df['triple'] + '<' + results_df['reference_results'] + '> '
    elif str(termtype).strip() == R2RML_LITERAL:
        results_df['reference_results'] = results_df['reference_results'].str.replace('"', '\\"', regex=False).str.replace('\\', '\\\\"', regex=False)
        results_df['triple'] = results_df['triple'] + '"' + results_df['reference_results'] + '"'
        if pd.notna(language_tag):
            results_df['triple'] = results_df['triple'] + '@' + language_tag + ' '
        elif pd.notna(datatype):
            results_df['triple'] = results_df['triple'] + '^^<' + datatype + '> '
        else:
            results_df['triple'] = results_df['triple'] + ' '

    return results_df


def _materialize_constant(results_df, constant, termtype=R2RML_IRI, language_tag='', datatype=''):
    if str(termtype).strip() == R2RML_LITERAL:
        complete_constant = '"' + constant + '"'

        if pd.notna(language_tag):
            complete_constant = complete_constant + '@' + language_tag + ' '
        elif pd.notna(datatype):
            complete_constant = complete_constant + '^^<' + datatype + '> '
        else:
            complete_constant = complete_constant + ' '
    else:
        complete_constant = '<' + str(constant) + '> '

    results_df['triple'] = results_df['triple'] + complete_constant

    return results_df


def _materialize_join_mapping_rule_terms(results_df, mapping_rule, parent_triples_map_rule, config):
    results_df['triple'] = ''
    if pd.notna(mapping_rule['subject_template']):
        results_df = _materialize_template(results_df, mapping_rule['subject_template'], config, termtype=mapping_rule['subject_termtype'], columns_alias='child_')
    elif pd.notna(mapping_rule['subject_constant']):
        results_df = _materialize_constant(results_df, mapping_rule['subject_constant'], termtype=mapping_rule['subject_termtype'])
    elif pd.notna(mapping_rule['subject_reference']):
        results_df = _materialize_reference(results_df, mapping_rule['subject_reference'], config, termtype=mapping_rule['subject_termtype'], columns_alias='child_')
    if pd.notna(mapping_rule['predicate_template']):
        results_df = _materialize_template(results_df, mapping_rule['predicate_template'], config, columns_alias='child_')
    elif pd.notna(mapping_rule['predicate_constant']):
        results_df = _materialize_constant(results_df, mapping_rule['predicate_constant'])
    elif pd.notna(mapping_rule['predicate_reference']):
        results_df = _materialize_reference(results_df, mapping_rule['predicate_reference'], config, termtype=R2RML_IRI, columns_alias='child_')
    if pd.notna(parent_triples_map_rule['subject_template']):
        results_df = _materialize_template(results_df, parent_triples_map_rule['subject_template'], config, termtype=parent_triples_map_rule['subject_termtype'], columns_alias='parent_')
    elif pd.notna(parent_triples_map_rule['subject_constant']):
        results_df = _materialize_constant(results_df, parent_triples_map_rule['subject_constant'], termtype=parent_triples_map_rule['subject_termtype'])
    elif pd.notna(parent_triples_map_rule['subject_reference']):
        results_df = _materialize_reference(results_df, parent_triples_map_rule['subject_reference'], config, termtype=parent_triples_map_rule['subject_termtype'], columns_alias='parent_')
    if pd.notna(mapping_rule['graph_template']):
        results_df = _materialize_template(results_df, mapping_rule['graph_template'], config, columns_alias='child_')
    elif pd.notna(mapping_rule['graph_constant']):
        if pd.notna(mapping_rule['graph_constant'] != R2RML_DEFAULT_GRAPH):
            results_df = _materialize_constant(results_df, mapping_rule['graph_constant'])
    elif pd.notna(mapping_rule['graph_reference']):
        results_df = _materialize_reference(results_df, mapping_rule['graph_reference'], config, termtype=R2RML_IRI, columns_alias='child_')

    return set(results_df['triple'])


def _materialize_mapping_rule_terms(results_df, mapping_rule, config):
    results_df['triple'] = ''
    if pd.notna(mapping_rule['subject_template']):
        results_df = _materialize_template(results_df, mapping_rule['subject_template'], config, termtype=mapping_rule['subject_termtype'])
    elif pd.notna(mapping_rule['subject_constant']):
        results_df = _materialize_constant(results_df, mapping_rule['subject_constant'], termtype=mapping_rule['subject_termtype'])
    elif pd.notna(mapping_rule['subject_reference']):
        results_df = _materialize_reference(results_df, mapping_rule['subject_reference'], config, termtype=mapping_rule['subject_termtype'])
    if pd.notna(mapping_rule['predicate_template']):
        results_df = _materialize_template(results_df, mapping_rule['predicate_template'])
    elif pd.notna(mapping_rule['predicate_constant']):
        results_df = _materialize_constant(results_df, mapping_rule['predicate_constant'])
    elif pd.notna(mapping_rule['predicate_reference']):
        results_df = _materialize_reference(results_df, mapping_rule['predicate_reference'], config, termtype=R2RML_IRI)
    if pd.notna(mapping_rule['object_template']):
        results_df = _materialize_template(results_df, mapping_rule['object_template'], config, termtype=mapping_rule['object_termtype'], language_tag=mapping_rule['object_language'], datatype=mapping_rule['object_datatype'])
    elif pd.notna(mapping_rule['object_constant']):
        results_df = _materialize_constant(results_df, mapping_rule['object_constant'], termtype=mapping_rule['object_termtype'], language_tag=mapping_rule['object_language'], datatype=mapping_rule['object_datatype'])
    elif pd.notna(mapping_rule['object_reference']):
        results_df = _materialize_reference(results_df, mapping_rule['object_reference'], config, termtype=mapping_rule['object_termtype'], language_tag=mapping_rule['object_language'], datatype=mapping_rule['object_datatype'])
    if config.get_output_format() == 'N-QUADS':
        if pd.notna(mapping_rule['graph_template']):
            results_df = _materialize_template(results_df, mapping_rule['graph_template'], config)
        elif pd.notna(mapping_rule['graph_constant']):
            if config.materialize_default_graph() or mapping_rule['graph_constant'] != R2RML_DEFAULT_GRAPH:
                results_df = _materialize_constant(results_df, mapping_rule['graph_constant'])
        elif pd.notna(mapping_rule['graph_reference']):
            results_df = _materialize_reference(results_df, mapping_rule['graph_reference'], config, termtype=R2RML_IRI)

    return set(results_df['triple'])


def _materalize_push_down_sql_join(mapping_rule, parent_triples_map_rule, references, parent_references, config):
    for key, join_condition in eval(mapping_rule['join_conditions']).items():
        parent_references.add(join_condition['parent_value'])
        references.add(join_condition['child_value'])

    triples_rule = set()
    result_chunks = get_sql_data(config, mapping_rule, references, parent_triples_map_rule, parent_references)
    for query_results_chunk_df in result_chunks:
        query_results_chunk_df = dataframe_columns_to_str(query_results_chunk_df)
        # query_results_chunk_df.dropna(axis=0, how='any', inplace=True)
        triples_rule.update(
            _materialize_join_mapping_rule_terms(query_results_chunk_df, mapping_rule, parent_triples_map_rule, config))

    return triples_rule


def _merge_results_chunks(query_results_chunk_df, parent_query_results_chunk_df, mapping_rule):
    child_join_references, parent_join_references = get_references_in_join_condition(mapping_rule)

    child_join_references = ['child_' + reference for reference in child_join_references]
    parent_join_references = ['parent_' + reference for reference in parent_join_references]

    return query_results_chunk_df.merge(parent_query_results_chunk_df,
                                        how='inner',
                                        left_on=child_join_references,
                                        right_on=parent_join_references)


def _materialize_mapping_rule(mapping_rule, subject_maps_df, config):
    triples = set()

    result_chunks = None
    parent_result_chunks = None

    references = _get_references_in_mapping_rule(mapping_rule)

    if pd.notna(mapping_rule['object_parent_triples_map']):
        parent_triples_map_rule = \
            subject_maps_df[subject_maps_df.triples_map_id == mapping_rule['object_parent_triples_map']].iloc[0]
        parent_references = _get_references_in_mapping_rule(parent_triples_map_rule, only_subject_map=True)

        if config.push_down_sql_joins() and \
                mapping_rule['source_type'] == RDB_SOURCE_TYPE and \
                mapping_rule['source_name'] == parent_triples_map_rule['source_name']:
            triples.update(_materalize_push_down_sql_join(mapping_rule, parent_triples_map_rule, references,
                                                          parent_references, config))

        else:
            # add references used in the join condition (not needed when pushing down join to SQL)
            references, parent_references = add_references_in_join_condition(mapping_rule, references,
                                                                                   parent_references)

            if mapping_rule['source_type'] == RDB_SOURCE_TYPE:
                result_chunks = get_sql_data(config, mapping_rule, references)
            elif mapping_rule['source_type'] in TABULAR_SOURCE_TYPES:
                result_chunks = get_table_data(config, mapping_rule, references)

            for query_results_chunk_df in result_chunks:
                query_results_chunk_df = dataframe_columns_to_str(query_results_chunk_df)
                #query_results_chunk_df.replace(config.get_na_values(), np.NaN)
                query_results_chunk_df.dropna(axis=0, how='any', inplace=True)
                query_results_chunk_df = query_results_chunk_df.add_prefix('child_')

                if parent_triples_map_rule['source_type'] == RDB_SOURCE_TYPE:
                    parent_result_chunks = get_sql_data(config, parent_triples_map_rule, parent_references)
                elif parent_triples_map_rule['source_type'] in TABULAR_SOURCE_TYPES:
                    parent_result_chunks = get_table_data(config, parent_triples_map_rule, parent_references)

                for parent_query_results_chunk_df in parent_result_chunks:
                    parent_query_results_chunk_df = dataframe_columns_to_str(parent_query_results_chunk_df)
                    #parent_query_results_chunk_df.replace(config.get_na_values(), np.NaN)
                    parent_query_results_chunk_df.dropna(axis=0, how='any', inplace=True)
                    parent_query_results_chunk_df = parent_query_results_chunk_df.add_prefix('parent_')
                    merged_query_results_chunk_df = _merge_results_chunks(query_results_chunk_df,
                                                                          parent_query_results_chunk_df, mapping_rule)

                    triples.update(
                        _materialize_join_mapping_rule_terms(merged_query_results_chunk_df, mapping_rule,
                                                             parent_triples_map_rule))

    else:
        if mapping_rule['source_type'] in RDB_SOURCE_TYPE:
            result_chunks = get_sql_data(config, mapping_rule, references)
        elif mapping_rule['source_type'] in TABULAR_SOURCE_TYPES:
            result_chunks = get_table_data(config, mapping_rule, references)

        for query_results_chunk_df in result_chunks:
            #query_results_chunk_df.replace(config.get_na_values(), np.NaN)
            query_results_chunk_df.dropna(axis=0, how='any', inplace=True)
            query_results_chunk_df = dataframe_columns_to_str(query_results_chunk_df)
            triples.update(_materialize_mapping_rule_terms(query_results_chunk_df, mapping_rule, config))

    return triples


def _materialize_mapping_partition(mapping_partition, subject_maps_df, config):
    triples = set()
    for i, mapping_rule in mapping_partition.iterrows():
        start_time = time.time()
        triples.update(set(_materialize_mapping_rule(mapping_rule, subject_maps_df, config)))

        logging.debug(str(len(triples)) + ' triples generated for mapping rule `' + str(
            mapping_rule['id']) + '` in ' + get_delta_time(start_time) + ' seconds.')

    triples_to_file(triples, config, mapping_partition.iloc[0]['mapping_partition'])

    return len(triples)


class Materializer:

    def __init__(self, mappings_df, config):
        self.mappings_df = mappings_df
        self.config = config

        self.subject_maps_df = get_subject_maps(mappings_df)
        self.mapping_partitions = [group for _, group in mappings_df.groupby(by='mapping_partition')]

        clean_output_dir(config)

    def __str__(self):
        return str(self.mappings_df)

    def __repr__(self):
        return repr(self.mappings_df)

    def __len__(self):
        return len(self.mappings_df)

    def materialize(self):
        num_triples = 0
        for mapping_partition in self.mapping_partitions:
            num_triples += _materialize_mapping_partition(mapping_partition, self.subject_maps_df, self.config)

        logging.info('Number of triples generated in total: ' + str(num_triples) + '.')

    def materialize_concurrently(self):
        logging.debug('Parallelizing with ' + str(self.config.get_number_of_processes()) + ' cores.')

        pool = mp.Pool(self.config.get_number_of_processes())
        num_triples = sum(pool.starmap(_materialize_mapping_partition,
                                       zip(self.mapping_partitions, repeat(self.subject_maps_df), repeat(self.config))))
        pool.close()
        pool.join()

        logging.info('Number of triples generated in total: ' + str(num_triples) + '.')



__author__ = "Julián Arenas-Guerrero"
__credits__ = ["Julián Arenas-Guerrero"]

__license__ = "Apache-2.0"
__maintainer__ = "Julián Arenas-Guerrero"
__email__ = "arenas.guerrero.julian@outlook.com"


import pandas as pd

from ..constants import *


def get_table_data(config, mapping_rule, references):
    tabular_source_type = mapping_rule['source_type']

    if tabular_source_type in [CSV_SOURCE_TYPE, TSV_SOURCE_TYPE]:
        return _read_csv(config, mapping_rule, references, tabular_source_type)
    elif tabular_source_type == EXCEL_SOURCE_TYPE:
        return _read_excel(config, mapping_rule, references)
    elif tabular_source_type == PARQUET_SOURCE_TYPE:
        return _read_parquet(mapping_rule, references)
    elif tabular_source_type == FEATHER_SOURCE_TYPE:
        return _read_feather(mapping_rule, references)
    elif tabular_source_type == ORC_SOURCE_TYPE:
        return _read_orc(mapping_rule, references)
    elif tabular_source_type == STATA_SOURCE_TYPE:
        return _read_stata(config, mapping_rule, references)
    elif tabular_source_type == SAS_SOURCE_TYPE:
        return _read_sas(config, mapping_rule)
    elif tabular_source_type == SPSS_SOURCE_TYPE:
        return _read_spss(mapping_rule, references)
    else:
        raise ValueError('Found an invalid source type. Found value `' + tabular_source_type + '`.')


def _read_csv(config, mapping_rule, references, tabular_source_type):
    delimiter = ',' if tabular_source_type == 'CSV' else '\t'

    return pd.read_table(mapping_rule['data_source'],
                         delimiter=delimiter,
                         index_col=False,   # TODO: use None?
                         usecols=references,
                         chunksize=config.get_chunksize(),
                         engine='c',
                         dtype=str,
                         memory_map=False,  # TODO: use True?
                         keep_default_na=False,
                         na_values=config.get_na_values(),
                         na_filter=config.apply_na_filter())


def _read_parquet(mapping_rule, references):
    parquet_df = pd.read_parquet(mapping_rule['data_source'],
                                 engine='pyarrow',
                                 columns=references)

    return [parquet_df]     # return as list because it does not support chunksize


def _read_feather(mapping_rule, references):
    feather_df = pd.read_feather(mapping_rule['data_source'],
                                 use_threads=True,
                                 columns=references)

    return [feather_df]


def _read_orc(mapping_rule, references):
    orc_df = pd.read_orc(mapping_rule['data_source'],
                         encoding='iso-8859-1',
                         columns=references)

    return [orc_df]


def _read_stata(config, mapping_rule, references):
    return pd.read_stata(mapping_rule['data_source'],
                         columns=references,
                         chunksize=config.get_chunksize(),
                         convert_dates=False,
                         convert_categoricals=False,
                         convert_missing=False,
                         preserve_dtypes=False,
                         order_categoricals=False)


def _read_sas(config, mapping_rule):
    return pd.read_sas(mapping_rule['data_source'],
                       encoding='iso-8859-1',
                       chunksize=config.get_chunksize())


def _read_spss(mapping_rule, references):
    spss_df = pd.read_spss(mapping_rule['data_source'],
                           usecols=references,
                           convert_categoricals=False)

    return [spss_df]


def _read_excel(config, mapping_rule, references):
    excel_df = pd.read_excel(mapping_rule['data_source'],
                            sheet_name=0,
                            engine='openpyxl',
                            usecols=references,
                            dtype=str,
                            keep_default_na=False,
                            na_values=config.get_na_values(),
                            na_filter=config.apply_na_filter())

    return [excel_df]
